unit SHA;

interface

uses
  Windows, KOL, WinSock,
    utils;

type
  pSHA1 = ^TSHA1;
  TSHA1 = object (TObj)
    private
      H0, H1, H2, H3, H4: ulong;
      bytes: array[0..63] of byte;
      unprocessedBytes: integer;
      size: ulong;
      procedure Process;
    public
      procedure AddBytes(data: pByte; num: integer);
      function HashFile(FileName: string): pByte;
      function GetDigest: pByte;

      destructor Destroy; virtual;
    end;

function NewSHA1: pSHA1;

implementation

function NewSHA1: pSHA1;
begin
  New(result, create);
  result.H0:=$67452301;
  result.H1:=$efcdab89;
  result.H2:=$98badcfe;
  result.H3:=$10325476;
  result.H4:=$c3d2e1f0;
  result.unprocessedBytes:=0;
  result.size:=0;
end;

function LRot(X: dword; c: integer): dword; register; assembler;
asm
  mov ecx, edx
  rol eax, cl
end;

procedure TSHA1.process;
var
  t, temp: integer;
  a, b, c, d, e, K, f: ulong;
  W: array[0..79] of ulong;
begin
  a:=H0;
  b:=H1;
  c:=H2;
  d:=H3;
  e:=H4;
  for t:=0 to 15 do
    W[t]:=(bytes[t*4] shl 24)+(bytes[t*4 + 1] shl 16)+(bytes[t*4 + 2] shl 8)+ bytes[t*4 + 3];
  for t:=t to 79 do
    W[t]:=lrot(W[t-3] xor W[t-8] xor W[t-14] xor W[t-16], 1);
  for t:=0 to 79 do
  begin
    if t<20 then
    begin
      k:=$5a827999;
      f:=(b and c) or ((b xor $FFFFFFFF) and d);
    end
      else if t<40 then
    begin
      k:=$6ed9eba1;
      f:=b xor c xor d;
    end
      else if t<60 then
    begin
      k:=$8f1bbcdc;
      f:=(b and c) or (b and d) or (c and d);
    end
      else
    begin
      k:=$ca62c1d6;
      f:=b xor c xor d;
    end;
    temp:=lrot(a, 5) + f + e + W[t] + K;
    e:=d;
    d:=c;
    c:=lrot(b, 30);
    b:=a;
    a:=temp;
  end;
  inc(H0, a);
  inc(H1, b);
  inc(H2, c);
  inc(H3, d);
  inc(H4, e);
  unprocessedBytes:=0;
end;

function TSHA1.HashFile(FileName: string): pByte;
var
  Stream: pStream;
  Data: pByte;
begin
  Stream:=NewReadFileStream(FileName);
  if Stream.Handle=INVALID_HANDLE_VALUE then
    Exit;
  GetMem(Data, Stream.Size);
  Stream.Read(Data^, Stream.Size);
  Self.AddBytes(Data, Stream.Size);
  FreeMem(Data);  
  result:=Self.GetDigest;
end;

procedure TSHA1.AddBytes(Data: pByte; Num: integer);
var
  needed, toCopy: integer;
begin
  inc(size, num);
  while num>0 do
  begin
    needed:=64-unprocessedBytes;
    if Num<Needed then toCopy:=Num
      else toCopy:=Needed;
    memcpy(@bytes[unprocessedBytes], data, toCopy);
    dec(Num, toCopy);
    inc(Data, toCopy);
    inc(unprocessedBytes, toCopy);
    if unprocessedBytes=64 then
      process;
  end;
end;

var
  footer: array[0..63] of byte = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);

function TSHA1.GetDigest: pByte;
var
  totalBitsL, totalBitsH: ulong;
  neededZeros: integer;
  digest, d: pByte;
  b: byte;
begin
  totalBitsL:=size shl 3;
  totalBitsH:=size shr 29;
  b:=$80;
  addBytes(@b, 1);
  if unprocessedBytes>56 then
    addBytes(@footer[0], 64-unprocessedBytes);
  neededZeros:=56-unprocessedBytes;     
  totalBitsH:=htonl(totalBitsH);
  totalBitsL:=htonl(totalBitsL);
  Move(totalBitsH, footer[neededZeros], 4);
  Move(totalBitsL, footer[neededZeros+4], 4);
  addBytes(@footer[0], neededZeros+8);
  GetMem(digest, 20);
  d:=digest;
  H0:=htonl(H0);
  H1:=htonl(H1);
  H2:=htonl(H2);
  H3:=htonl(H3);
  H4:=htonl(H4);
  Move(H0, d^, 4);  inc(d, 4);
  Move(H1, d^, 4);  inc(d, 4);
  Move(H2, d^, 4);  inc(d, 4);
  Move(H3, d^, 4);  inc(d, 4);
  Move(H4, d^, 4);  inc(d, 4);
  result:=digest;
end;

destructor TSHA1.Destroy;
begin
end;


end.
